@@ -1,40 +1,32 @@
-/* Elliptic Curve Method implementation.
+/* Elliptic Curve Method: toplevel and stage 1 routines.
 
-  Copyright 2001, 2002, 2003 Paul Zimmermann and Alexander Kruppa.
+  Copyright 2001, 2002, 2003, 2004, 2005 Paul Zimmermann and Alexander Kruppa.
 
-  This program is free software; you can redistribute it and/or modify it
-  under the terms of the GNU General Public License as published by the
-  Free Software Foundation; either version 2 of the License, or (at your
+  This file is part of the ECM Library.
+
+  The ECM Library is free software; you can redistribute it and/or modify
+  it under the terms of the GNU Lesser General Public License as published by
+  the Free Software Foundation; either version 2.1 of the License, or (at your
   option) any later version.
 
-  This program is distributed in the hope that it will be useful, but WITHOUT
-  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
-  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
-  more details.
+  The ECM Library is distributed in the hope that it will be useful, but
+  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
+  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
+  License for more details.
 
-  You should have received a copy of the GNU General Public License along
-  with this program; see the file COPYING.  If not, write to the Free
-  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-  02111-1307, USA.
+  You should have received a copy of the GNU Lesser General Public License
+  along with the ECM Library; see the file COPYING.LIB.  If not, write to
+  the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
+  MA 02111-1307, USA.
 */
 
 #include <stdio.h>
 #include <stdlib.h>
 #include <math.h>
+#include <limits.h> /* for ULONG_MAX */
 #include "gmp.h"
 #include "ecm.h"
-
-int get_curve_from_sigma (mpz_t, mpres_t, mpres_t, mpres_t, mpmod_t);
-int montgomery_to_weierstrass (mpz_t, mpres_t, mpres_t, mpres_t, mpmod_t);
-void add3 (mpres_t, mpres_t, mpres_t, mpres_t, mpres_t, mpres_t, mpres_t,
-           mpres_t, mpmod_t, mpres_t, mpres_t, mpres_t);
-void duplicate (mpres_t, mpres_t, mpres_t, mpres_t, mpmod_t, mpres_t,
-                mpres_t, mpres_t, mpres_t);
-unsigned int lucas_cost (unsigned, double);
-void prac (mpres_t, mpres_t, unsigned int, mpmod_t, mpres_t, mpres_t, mpres_t, 
-           mpres_t, mpres_t, mpres_t, mpres_t, mpres_t, mpres_t, mpres_t, 
-           mpres_t, mpres_t);
-int ecm_stage1 (mpz_t, mpres_t, mpres_t, mpmod_t, double, double);
+#include "ecm-impl.h"
 
 /******************************************************************************
 *                                                                             *
@@ -44,13 +36,12 @@ int ecm_stage1 (mpz_t, mpres_t, mpres_t, mpmod_t, double, double);
 
 #define mpz_mulmod5(r,s1,s2,m,t) { mpz_mul(t,s1,s2); mpz_mod(r, t, m); }
 
-
 /* Computes curve parameter A and a starting point (x:1) from a given 
    sigma value.
    If a factor of n was found during the process, returns 1 (and factor 
    in A), 0 otherwise.
 */
-int
+static int
 get_curve_from_sigma (mpz_t f, mpres_t A, mpres_t x, mpz_t sigma, mpmod_t n)
 {
   mpres_t t, u, v, b, z;
@@ -113,7 +104,7 @@ get_curve_from_sigma (mpz_t f, mpres_t A, mpres_t x, mpz_t sigma, mpmod_t n)
    by change of variables x -> g*X-a/3, y -> g*Y.
    We have A = (3-a^2)/(3g^2), X = (3x+a)/(3g), Y = y/g.
 */
-int 
+static int 
 montgomery_to_weierstrass (mpz_t f, mpres_t x, mpres_t y, mpres_t A, mpmod_t n)
 {
   mpres_t g;
@@ -152,16 +143,15 @@ montgomery_to_weierstrass (mpz_t f, mpres_t x, mpres_t y, mpres_t A, mpmod_t n)
   return 0;
 }
 
-
 /* adds Q=(x2:z2) and R=(x1:z1) and puts the result in (x3:z3),
      using 6 muls (4 muls and 2 squares), and 6 add/sub.
    One assumes that Q-R=P or R-Q=P where P=(x:z).
      - n : number to factor
-     - t, u, v, w : auxiliary variables
-   Modifies: x3, z3, t, u, v, w.
+     - u, v, w : auxiliary variables
+   Modifies: x3, z3, u, v, w.
    (x3,z3) may be identical to (x2,z2) and to (x,z)
 */
-void
+static void
 add3 (mpres_t x3, mpres_t z3, mpres_t x2, mpres_t z2, mpres_t x1, mpres_t z1, 
       mpres_t x, mpres_t z, mpmod_t n, mpres_t u, mpres_t v, mpres_t w)
 {
@@ -203,7 +193,7 @@ add3 (mpres_t x3, mpres_t z3, mpres_t x2, mpres_t z2, mpres_t x1, mpres_t z1,
      - b : (a+2)/4 mod n
      - t, u, v, w : auxiliary variables
 */
-void
+static void
 duplicate (mpres_t x2, mpres_t z2, mpres_t x1, mpres_t z1, mpmod_t n, 
            mpres_t b, mpres_t u, mpres_t v, mpres_t w)
 {
@@ -218,8 +208,62 @@ duplicate (mpres_t x2, mpres_t z2, mpres_t x1, mpres_t z1, mpmod_t n,
   mpres_mul (z2, w, u, n);  /* z2 = ((4*x1*z1)*((x1-z1)^2+(A+2)/4*(4*x1*z1))) mod n */
 }
 
-#define ADD 6 /* number of multiplications in an addition */
-#define DUP 5 /* number of multiplications in a duplicate */
+/* multiply P=(x:z) by e and puts the result in (x:z).
+   Assumes e >= 1.
+*/
+static void
+ecm_mul (mpres_t x, mpres_t z, mpz_t e, mpmod_t n, mpres_t b)
+{
+  size_t l;
+  mpres_t x0, z0, x1, z1, u, v, w;
+
+  if (mpz_cmp_ui (e, 1) == 0)
+    return;
+
+  mpres_init (x0, n);
+  mpres_init (z0, n);
+  mpres_init (x1, n);
+  mpres_init (z1, n);
+  mpres_init (u, n);
+  mpres_init (v, n);
+  mpres_init (w, n);
+
+  l = mpz_sizeinbase (e, 2) - 1; /* l >= 1 */
+
+  mpres_set (x0, x, n);
+  mpres_set (z0, z, n);
+  duplicate (x1, z1, x0, z0, n, b, u, v, w);
+
+  /* invariant: (P1,P0) = ((k+1)P, kP) where k = floor(e/2^l) */
+
+  while (l-- > 0)
+    {
+      if (mpz_tstbit (e, l)) /* k, k+1 -> 2k+1, 2k+2 */
+        {
+          add3 (x0, z0, x0, z0, x1, z1, x, z, n, u, v, w); /* 2k+1 */
+          duplicate (x1, z1, x1, z1, n, b, u, v, w); /* 2k+2 */
+        }
+      else /* k, k+1 -> 2k, 2k+1 */
+        {
+          add3 (x1, z1, x1, z1, x0, z0, x, z, n, u, v, w); /* 2k+1 */
+          duplicate (x0, z0, x0, z0, n, b, u, v, w); /* 2k */
+        }
+    }
+
+  mpres_set (x, x0, n);
+  mpres_set (z, z0, n);
+
+  mpres_clear (x0, n);
+  mpres_clear (z0, n);
+  mpres_clear (x1, n);
+  mpres_clear (z1, n);
+  mpres_clear (u, n);
+  mpres_clear (v, n);
+  mpres_clear (w, n);
+}
+
+#define ADD 6.0 /* number of multiplications in an addition */
+#define DUP 5.0 /* number of multiplications in a duplicate */
 
 #define START(d,v) ((d)/1.6180339887498948482-128.0+(v))
 
@@ -228,15 +272,16 @@ duplicate (mpres_t x2, mpres_t z2, mpres_t x1, mpres_t z1, mpmod_t n,
    ADD is the cost of an addition
    DUP is the cost of a duplicate
 */
-unsigned int
-lucas_cost (unsigned n, double v)
+static double
+lucas_cost (unsigned long n, double v)
 {
-  unsigned int c, d, e, r;
+  unsigned long d, e, r;
+  double c; /* cost */
 
   d = n;
-  r = (unsigned int) ((double) d / v + 0.5);
+  r = (unsigned long) ((double) d / v + 0.5);
   if (r >= n)
-    return (ADD * n);
+    return (ADD * (double) n);
   d = n - r;
   e = 2 * r - n;
   c = DUP + ADD; /* initial duplicate and final addition */
@@ -252,12 +297,12 @@ lucas_cost (unsigned n, double v)
         { /* condition 1 */
           d = (2 * d - e) / 3;
           e = (e - d) / 2;
-          c += 3 * ADD; /* 3 additions */
+          c += 3.0 * ADD; /* 3 additions */
         }
       else if (4 * d <= 5 * e && (d - e) % 6 == 0)
         { /* condition 2 */
           d = (d - e) / 2;
-          c += ADD+DUP; /* one addition, one duplicate */
+          c += ADD + DUP; /* one addition, one duplicate */
         }
       else if (d <= 4 * e)
         { /* condition 3 */
@@ -268,70 +313,67 @@ lucas_cost (unsigned n, double v)
         { /* condition 4 */
           d = (d - e) / 2;
           c += ADD + DUP; /* one addition, one duplicate */
-        } 
+        }
+      /* now d+e is odd */
       else if (d % 2 == 0)
         { /* condition 5 */
           d /= 2;
-          c += ADD+DUP; /* one addition, one duplicate */
+          c += ADD + DUP; /* one addition, one duplicate */
         }
+      /* now d is odd and e is even */
       else if (d % 3 == 0)
         { /* condition 6 */
           d = d / 3 - e;
-          c += 3 * ADD + DUP; /* three additions, one duplicate */
+          c += 3.0 * ADD + DUP; /* three additions, one duplicate */
         }
       else if ((d + e) % 3 == 0)
         { /* condition 7 */
           d = (d - 2 * e) / 3;
-          c += 3 * ADD + DUP; /* three additions, one duplicate */
+          c += 3.0 * ADD + DUP; /* three additions, one duplicate */
         }
       else if ((d - e) % 3 == 0)
         { /* condition 8 */
           d = (d - e) / 3;
-          c += 3 * ADD + DUP; /* three additions, one duplicate */
+          c += 3.0 * ADD + DUP; /* three additions, one duplicate */
         }
-      else if (e % 2 == 0)
+      else /* necessarily e is even: catches all cases */
         { /* condition 9 */
           e /= 2;
           c += ADD + DUP; /* one addition, one duplicate */
         }
-      else
-        {
-          fprintf (stderr, "lucas_cost: no condition qualifies for d=%u e=%u\n", d, e);
-          exit (EXIT_FAILURE);
-        }
     }
   
-  return (c);
+  return c;
 }
 
 
-#define NV 10
-
 /* computes kP from P=(xA:zA) and puts the result in (xA:zA). Assumes k>2. */
-void
-prac (mpres_t xA, mpres_t zA, unsigned int k, mpmod_t n, mpres_t b, 
+static void
+prac (mpres_t xA, mpres_t zA, unsigned long k, mpmod_t n, mpres_t b,
       mpres_t u, mpres_t v, mpres_t w, mpres_t xB, mpres_t zB, mpres_t xC, 
       mpres_t zC, mpres_t xT, mpres_t zT, mpres_t xT2, mpres_t zT2)
 {
-  unsigned int d, e, r, i = 0;
+  unsigned long d, e, r, i = 0;
+  double c, cmin;
   __mpz_struct *tmp;
+#define NV 10  
   static double val[NV] =
     { 1.61803398875, 1.72360679775, 1.618347119656, 1.617914406529,
       1.612429949509, 1.632839806089, 1.620181980807, 1.580178728295,
       1.617214616534, 1.38196601125 };
   
   /* chooses the best value of v */
-  for (d = 0, r = ADD * k; d < NV; d++)
+  for (d = 0, cmin = ADD * (double) k; d < NV; d++)
     {
-      e = lucas_cost (k, val[d]);
-      if (e < r)
+      c = lucas_cost (k, val[d]);
+      if (c < cmin)
         {
-          r = e;
+          cmin = c;
           i = d;
         }
     }
   d = k;
-  r = (int) ((double) d / val[i] + 0.5);
+  r = (unsigned long) ((double) d / val[i] + 0.5);
   
   /* first iteration always begins by Condition 3, then a swap */
   d = k - r;
@@ -388,19 +430,21 @@ prac (mpres_t xA, mpres_t zA, unsigned int k, mpmod_t n, mpres_t b,
           add3 (xB, zB, xB, zB, xA, zA, xC, zC, n, u, v, w); /* B = f(B,A,C) */
           duplicate (xA, zA, xA, zA, n, b, u, v, w); /* A = 2*A */
         }
+      /* now d+e is odd */
       else if (d % 2 == 0)
         { /* condition 5 */
           d /= 2;
           add3 (xC, zC, xC, zC, xA, zA, xB, zB, n, u, v, w); /* C = f(C,A,B) */
           duplicate (xA, zA, xA, zA, n, b, u, v, w); /* A = 2*A */
         }
+      /* now d is odd, e is even */
       else if (d % 3 == 0)
         { /* condition 6 */
           d = d / 3 - e;
-          duplicate (xT, zT, xA, zA, n, b, u, v, w); /* T1 = 2*A */
+          duplicate (xT, zT, xA, zA, n, b, u, v, w); /* T = 2*A */
           add3 (xT2, zT2, xA, zA, xB, zB, xC, zC, n, u, v, w); /* T2 = f(A,B,C) */
-          add3 (xA, zA, xT, zT, xA, zA, xA, zA, n, u, v, w); /* A = f(T1,A,A) */
-          add3 (xT, zT, xT, zT, xT2, zT2, xC, zC, n, u, v, w); /* T1 = f(T1,T2,C) */
+          add3 (xA, zA, xT, zT, xA, zA, xA, zA, n, u, v, w); /* A = f(T,A,A) */
+          add3 (xT, zT, xT, zT, xT2, zT2, xC, zC, n, u, v, w); /* T = f(T,T2,C) */
           /* circular permutation (C,B,T) */
           tmp = xC;
           xC = xB;
@@ -414,43 +458,32 @@ prac (mpres_t xA, mpres_t zA, unsigned int k, mpmod_t n, mpres_t b,
       else if ((d + e) % 3 == 0)
         { /* condition 7 */
           d = (d - 2 * e) / 3;
-          add3 (xT, zT, xA, zA, xB, zB, xC, zC, n, u, v, w); /* T1 = f(A,B,C) */
-          add3 (xB, zB, xT, zT, xA, zA, xB, zB, n, u, v, w); /* B = f(T1,A,B) */
+          add3 (xT, zT, xA, zA, xB, zB, xC, zC, n, u, v, w); /* T = f(A,B,C) */
+          add3 (xB, zB, xT, zT, xA, zA, xB, zB, n, u, v, w); /* B = f(T,A,B) */
           duplicate (xT, zT, xA, zA, n, b, u, v, w);
           add3 (xA, zA, xA, zA, xT, zT, xA, zA, n, u, v, w); /* A = 3*A */
         }
       else if ((d - e) % 3 == 0)
         { /* condition 8 */
           d = (d - e) / 3;
-          add3 (xT, zT, xA, zA, xB, zB, xC, zC, n, u, v, w); /* T1 = f(A,B,C) */
+          add3 (xT, zT, xA, zA, xB, zB, xC, zC, n, u, v, w); /* T = f(A,B,C) */
           add3 (xC, zC, xC, zC, xA, zA, xB, zB, n, u, v, w); /* C = f(A,C,B) */
           mpres_swap (xB, xT, n);
           mpres_swap (zB, zT, n); /* swap B and T */
           duplicate (xT, zT, xA, zA, n, b, u, v, w);
           add3 (xA, zA, xA, zA, xT, zT, xA, zA, n, u, v, w); /* A = 3*A */
         }
-      else if (e % 2 == 0)
+      else /* necessarily e is even here */
         { /* condition 9 */
           e /= 2;
           add3 (xC, zC, xC, zC, xB, zB, xA, zA, n, u, v, w); /* C = f(C,B,A) */
           duplicate (xB, zB, xB, zB, n, b, u, v, w); /* B = 2*B */
         }
-      else
-        {
-          fprintf (stderr, "no condition qualifies for d=%u e=%u\n", d, e);
-          exit (EXIT_FAILURE);
-        }
     }
   
   add3 (xA, zA, xA, zA, xB, zB, xC, zC, n, u, v, w);
-  
-#ifdef DEBUG
-  if (d != 1)
-    {
-      fprintf (stderr, "d!=1 at the end of PRAC\n");
-      exit (EXIT_FAILURE);
-    }
-#endif
+
+  ASSERT(d == 1);
 }
 
 
@@ -459,17 +492,14 @@ prac (mpres_t xA, mpres_t zA, unsigned int k, mpmod_t n, mpres_t b,
           g*y^2*z = x^3 + a*x^2*z + x*z^2
           n is the number to factor
 	  B1 is the stage 1 bound
-	  B2 is the stage 2 bound
-          verbose is verbosity level: 0 no output, 1 normal output,
-            2 diagnostic output
    Output: If a factor is found, it is returned in x.
            Otherwise, x contains the x-coordinate of the point computed
            in stage 1 (with z coordinate normalized to 1).
    Return value: non-zero iff a factor is found
 */
-int
-ecm_stage1 (mpz_t f, mpres_t x, mpres_t A, mpmod_t n, double B1, 
-            double B1done)
+static int
+ecm_stage1 (mpz_t f, mpres_t x, mpres_t A, mpmod_t n, double B1, double B1done,
+	    mpz_t go)
 {
   mpres_t b, z, u, v, w, xB, zB, xC, zC, xT, zT, xT2, zT2;
   double q, r;
@@ -488,12 +518,16 @@ ecm_stage1 (mpz_t f, mpres_t x, mpres_t A, mpmod_t n, double B1,
   mpres_init (zT, n);
   mpres_init (xT2, n);
   mpres_init (zT2, n);
-  
+
   mpres_set_ui (z, 1, n);
   
   mpres_add_ui (b, A, 2, n);
   mpres_div_2exp (b, b, 2, n); /* b == (A+2)/4 */
 
+  /* preload group order */
+  if (go != NULL)
+    ecm_mul (x, z, go, n, b);
+
   /* prac() wants multiplicands > 2 */
   for (r = 2.0; r <= B1; r *= 2.0)
     if (r > B1done)
@@ -509,11 +543,13 @@ ecm_stage1 (mpz_t f, mpres_t x, mpres_t A, mpmod_t n, double B1,
   
   q = getprime (2.0); /* Puts 3.0 into q. Next call gives 5.0 */
   for (q = getprime (q); q <= B1; q = getprime (q))
-    for (r = q; r <= B1; r *= q)
-      if (r > B1done)
-	  prac (x, z, (int) q, n, b, u, v, w, xB, zB, xC, zC, xT, zT, xT2, zT2);
-  getprime (0.0); /* free the prime tables, and reinitialize */
-
+    {
+      for (r = q; r <= B1; r *= q)
+	if (r > B1done)
+	  prac (x, z, (unsigned long) q, n, b, u, v, w, xB, zB, xC, zC, xT,
+		zT, xT2, zT2);
+    }
+  getprime (FREE_PRIME_TABLE); /* free the prime tables, and reinitialize */
 
   /* Normalize z to 1 */
   if (!mpres_invert (u, z, n)) /* Factor found? */
@@ -536,34 +572,76 @@ ecm_stage1 (mpz_t f, mpres_t x, mpres_t A, mpmod_t n, double B1,
   mpres_clear (u, n);
   mpres_clear (z, n);
   mpres_clear (b, n);
-  
+
   return ret;
 }
 
+/* choose "optimal" S according to step 2 range B2 */
+int
+choose_S (mpz_t B2len)
+{
+  if (mpz_cmp_d (B2len, 1e7) < 0)
+    return 1;   /* x^1 */
+  else if (mpz_cmp_d (B2len, 1e8) < 0)
+    return 2;   /* x^2 */
+  else if (mpz_cmp_d (B2len, 1e9) < 0)
+    return -3;  /* Dickson(3) */
+  else if (mpz_cmp_d (B2len, 1e10) < 0)
+    return -6;  /* Dickson(6) */
+  else if (mpz_cmp_d (B2len, 3e11) < 0)
+    return -12; /* Dickson(12) */
+  else
+    return -30; /* Dickson(30) */
+}
+
 /* Input: x is starting point or zero
           sigma is sigma value (if x is set to zero) or 
             A parameter (if x is non-zero) of curve
           n is the number to factor
+          go is the initial group order to preload  
           B1, B2 are the stage 1/stage 2 bounds, respectively
+          B2min the lower bound for stage 2
+          B2scale is the stage 2 scale factor
           k is the number of blocks to do in stage 2
           S is the degree of the Suyama-Brent extension for stage 2
           verbose is verbosity level: 0 no output, 1 normal output,
             2 diagnostic output.
           sigma_is_a: If true, the sigma parameter contains the curve's A value
    Output: f is the factor found.
-   Return value: non-zero iff a factor was found.
+   Return value: ECM_FACTOR_FOUND_STEPn if a factor was found,
+                 ECM_NO_FACTOR_FOUND if no factor was found,
+		 ECM_ERROR in case of error.
 */
 int
-ecm (mpz_t f, mpz_t x, mpz_t sigma, mpz_t n, double B1done, double B1,
-     double B2min, double B2, unsigned int k, int S, int verbose, int repr,
-     int sigma_is_A)
+ecm (mpz_t f, mpz_t x, mpz_t sigma, mpz_t n, mpz_t go, double B1done,
+     double B1, mpz_t B2min_parm, mpz_t B2_parm, double B2scale, 
+     unsigned int k, int S, int verbose, int repr, int sigma_is_A, 
+     FILE *os, FILE* es, char *TreeFilename)
 {
-  int youpi = 0, st;
+  int youpi = ECM_NO_FACTOR_FOUND;
+  unsigned int st;
   mpmod_t modulus;
   curve P;
+  mpz_t B2min, B2; /* Local B2, B2min to avoid changing caller's values */
+
+  set_verbose (verbose);
+  ECM_STDOUT = (os == NULL) ? stdout : os;
+  ECM_STDERR = (es == NULL) ? stdout : es;
+
+  /* if n is even, return 2 */
+  if (mpz_divisible_2exp_p (n, 1))
+    {
+      mpz_set_ui (f, 2);
+      return ECM_FACTOR_FOUND_STEP1;
+    }
+
+  /* now n is odd */
 
   st = cputime ();
 
+  mpz_init_set (B2min, B2min_parm);
+  mpz_init_set (B2, B2_parm);
+
   /* set second stage bound B2: when using polynomial multiplication of
      complexity n^alpha, stage 2 has complexity about B2^(alpha/2), and
      we want stage 2 to take about half of stage 1, thus we choose
@@ -576,9 +654,29 @@ ecm (mpz_t f, mpz_t x, mpz_t sigma, mpz_t n, double B1done, double B1,
      addition and duplicate, and both are optimized with PRAC, we can
      assume the ratio remains about 11/2. */
 
-  if (B2 == 0.0)
-    B2 = pow (11.0 / 6.0 * B1, 1.424828748);
-  
+  /* Also scale B2 by what the user said (or by the default scaling of 1.0) */
+
+  if (ECM_IS_DEFAULT_B2(B2))
+    mpz_set_d (B2, B2scale * pow (11.0 / 6.0 * B1, 1.424828748));
+
+  /* set B2min */
+  if (mpz_sgn (B2min) < 0)
+    mpz_set_d (B2min, B1);
+
+  if (repr == 1)
+    mpmod_init_MPZ (modulus, n);
+  else if (repr == MOD_MODMULN)
+    mpmod_init_MODMULN (modulus, n);
+  else if (repr == MOD_REDC)
+    mpmod_init_REDC (modulus, n);
+  else if (abs (repr) > 16)
+    {
+      if (mpmod_init_BASE2 (modulus, repr, n) == ECM_ERROR)
+        return ECM_ERROR;
+    }
+  else /* automatic choice, avoiding base2 if repr=-1 */
+    mpmod_init (modulus, n, repr);
+
   /* Set default degree for Brent-Suyama extension */
   /* We try to keep the time used by the Brent-Suyama extension
      at about 10% of the stage 2 time */
@@ -586,56 +684,53 @@ ecm (mpz_t f, mpz_t x, mpz_t sigma, mpz_t n, double B1done, double B1,
      the better Dickson polys whenever possible. For S == 1, 2, they behave
      identically. */
 
-  if (S == 0)
+  if (S == ECM_DEFAULT_S)
     {
-      if (B2 - B2min < 1.e7)
-        S = 1;    /* x^1 */
-      else if (B2 - B2min < 1.e8)
-        S = 2;   /* x^2 */
-      else if (B2 - B2min < 8.e8)
-        S = -3;   /* Dickson(3) */
-      else if (B2 - B2min < 4.e9)
-        S = -6;  /* Dickson(6) */
-      else if (B2 - B2min < 3.e10)
-        S = -12;  /* Dickson(12) */
+      /* This requires that modulus is already inited */
+      if (modulus->repr == MOD_BASE2 && modulus->Fermat > 0)
+        {
+          /* For Fermat numbers, default is 1 (no Brent-Suyama) */
+          S = 1;
+        }
       else
-        S = -30;  /* Dickson(30) */
+        {
+          mpz_t t;
+          mpz_init (t);
+          mpz_sub (t, B2, B2min);
+          S = choose_S (t);
+          mpz_clear (t);
+        }
     }
   
-  if (repr == 1)
-    mpmod_init_MPZ (modulus, n);
-  else if (repr == 2)
-    mpmod_init_MODMULN (modulus, n);
-  else if (repr == 3)
-    mpmod_init_REDC (modulus, n);
-  else if (repr > 16)
-    mpmod_init_BASE2 (modulus, repr, n);
-  else
-    mpmod_init (modulus, n, repr, verbose);
-
   mpres_init (P.x, modulus);
   mpres_init (P.A, modulus);
 
   if (sigma_is_A == 0)
     {
-      /* sigma contains sigma value, A value must be computed */
-      if (verbose >= 1)
+      /* if sigma=0, generate it at random */
+      if (mpz_cmp_ui (sigma, 0) == 0)
         {
-          printf ("Using B1=%1.0f, B2=", B1);
-	  if (B2min <= B1)
-	    printf ("%1.0f", B2);
-	  else
-	    printf ("%1.0f-%1.0f", B2min, B2);
-	  printf (", polynomial ");
-	  if (S > 0)
-	    printf ("x^%u", S);
-	  else
-	    printf ("Dickson(%u)", -S);
-	  printf (", sigma=");
-          mpz_out_str (stdout, 10, sigma);
-          printf ("\n");
-          fflush (stdout);
+          gmp_randstate_t state;
+          gmp_randinit_default (state);
+          gmp_randseed_ui (state, get_random_ui ());
+          mpz_urandomb (sigma, state, 32);
+          mpz_add_ui (sigma, sigma, 6);
+          gmp_randclear (state);
         }
+
+      /* sigma contains sigma value, A value must be computed */
+      outputf (OUTPUT_NORMAL, "Using B1=%1.0f, B2=", B1);
+      if (mpz_cmp_d (B2min, B1) == 0)
+        outputf (OUTPUT_NORMAL, "%Zd", B2);
+      else
+        outputf (OUTPUT_NORMAL, "%Zd-%Zd", B2min, B2);
+      outputf (OUTPUT_NORMAL, ", polynomial ");
+      if (S > 0)
+        outputf (OUTPUT_NORMAL, "x^%u", S);
+      else
+        outputf (OUTPUT_NORMAL, "Dickson(%u)", -S);
+      outputf (OUTPUT_NORMAL, ", sigma=%Zd\n", sigma);
+
       if ((youpi = get_curve_from_sigma (f, P.A, P.x, sigma, modulus)))
         goto end_of_ecm;
     }
@@ -647,8 +742,10 @@ ecm (mpz_t f, mpz_t x, mpz_t sigma, mpz_t n, double B1done, double B1,
       /* For now, we'll just chicken out. */
       if (mpz_sgn (x) == 0)
         {
-          fprintf (stderr, "The -A option requires that a starting point is given as well by the -x0\noption\n");
-          exit (EXIT_FAILURE);
+          outputf (OUTPUT_ERROR, 
+                   "Error, -A requires a starting point (-x0 x).\n");
+	  youpi = ECM_ERROR;
+	  goto end_of_ecm;
         }
     }
 
@@ -656,46 +753,62 @@ ecm (mpz_t f, mpz_t x, mpz_t sigma, mpz_t n, double B1done, double B1,
   if (mpz_sgn (x) != 0)
       mpres_set_z (P.x, x, modulus);
   
-  if (verbose >= 2)
+  if (test_verbose (OUTPUT_RESVERBOSE))
     {
-      printf ("a=");
-      mpres_out_str (stdout, 10, P.A, modulus);
-      printf ("\nstarting point: x=");
-      mpres_out_str (stdout, 10, P.x, modulus);
-      printf ("\n");
+      mpz_t t;
+      
+      mpz_init (t);
+      mpres_get_z (t, P.A, modulus);
+      outputf (OUTPUT_RESVERBOSE, "a=%Zd\n", t);
+      mpres_get_z (t, P.x, modulus);
+      outputf (OUTPUT_RESVERBOSE, "starting point: x=%Zd\n", t);
+      mpz_clear (t);
+    }
+
+  if (go != NULL && mpz_cmp_ui (go, 1) > 0)
+    outputf (OUTPUT_VERBOSE, "initial group order: %Zd\n", go);
+
+  /* check that B1 is not too large */
+  if (B1 > (double) ULONG_MAX)
+    {
+      outputf (OUTPUT_ERROR, "Error, maximal step1 bound for ECM is %lu.\n", 
+               ULONG_MAX);
+      youpi = ECM_ERROR;
+      goto end_of_ecm;
     }
 
   if (B1 > B1done)
-    youpi = ecm_stage1 (f, P.x, P.A, modulus, B1, B1done);
-
-  if (verbose >= 1)
-    {
-      printf ("Step 1 took %dms\n", cputime () - st);
-      fflush (stdout);
-    }
+    youpi = ecm_stage1 (f, P.x, P.A, modulus, B1, B1done, go);
+  
+  st = elltime (st, cputime ());
+  
+  outputf (OUTPUT_NORMAL, "Step 1 took %ums\n", st);
 
   /* Store end-of-stage-1 residue in x in case we write it to a save file, 
      before P.x is converted to Weierstrass form */
   
   mpres_get_z (x, P.x, modulus);
 
-  if (youpi != 0) /* a factor was found */
+  if (youpi != ECM_NO_FACTOR_FOUND) /* a factor was found */
     goto end_of_ecm;
 
-  if (verbose >= 2) 
+  if (test_verbose (OUTPUT_RESVERBOSE)) 
     {
-      printf ("x=");
-      mpres_out_str (stdout, 10, P.x, modulus);
-      printf ("\n");
-      fflush (stdout);
+      mpz_t t;
+      
+      mpz_init (t);
+      mpres_get_z (t, P.x, modulus);
+      outputf (OUTPUT_RESVERBOSE, "x=%Zd\n", t);
+      mpz_clear (t);
     }
 
   mpres_init (P.y, modulus);
 
   youpi = montgomery_to_weierstrass (f, P.x, P.y, P.A, modulus);
-  
-  if (youpi == 0)
-    youpi = stage2 (f, &P, modulus, B2min, B2, k, S, verbose, EC_METHOD);
+
+  if (youpi == ECM_NO_FACTOR_FOUND)
+    youpi = stage2 (f, &P, modulus, B2min, B2, k, S, ECM_ECM, st, 
+                    TreeFilename);
   
   mpres_clear (P.y, modulus);
 
@@ -703,6 +816,8 @@ end_of_ecm:
   mpres_clear (P.x, modulus);
   mpres_clear (P.A, modulus);
   mpmod_clear (modulus);
+  mpz_clear (B2);
+  mpz_clear (B2min);
 
   return youpi;
 }
