@@ -1,21 +1,23 @@
 /* Dynamic Eratosthenes sieve.
 
-  Copyright 2001, 2002, 2003 Paul Zimmermann and Alexander Kruppa.
+  Copyright 2001, 2002, 2003, 2005 Paul Zimmermann and Alexander Kruppa.
 
-  This program is free software; you can redistribute it and/or modify it
-  under the terms of the GNU General Public License as published by the
-  Free Software Foundation; either version 2 of the License, or (at your
+  This file is part of the ECM Library.
+
+  The ECM Library is free software; you can redistribute it and/or modify
+  it under the terms of the GNU Lesser General Public License as published by
+  the Free Software Foundation; either version 2.1 of the License, or (at your
   option) any later version.
 
-  This program is distributed in the hope that it will be useful, but WITHOUT
-  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
-  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
-  more details.
+  The ECM Library is distributed in the hope that it will be useful, but
+  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
+  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
+  License for more details.
 
-  You should have received a copy of the GNU General Public License along
-  with this program; see the file COPYING.  If not, write to the Free
-  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-  02111-1307, USA.
+  You should have received a copy of the GNU Lesser General Public License
+  along with the ECM Library; see the file COPYING.LIB.  If not, write to
+  the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
+  MA 02111-1307, USA.
 */
 
 #include <math.h>
@@ -23,23 +25,26 @@
 #include <stdlib.h>
 #include "gmp.h"
 #include "ecm.h"
-
-/*
-  B1          10^5    10^6     10^7     10^8        1e9        1e10
-  Pi(B1)      9592    78498    664579   5761455   50847534    455052511
-  prev(B1)    99991   999983   9999991  99999989  999999937   9999999967
-  ecm/step1   0.1     2.02     42.7     
-  getprime    0.0     0.04     0.38     3.87        43.64      536.8
-*/
+#ifdef OUTSIDE_LIBECM
+#include "ecm-ecm.h"
+#else
+#include "ecm-impl.h"
+#endif
 
 /* This function returns successive odd primes, starting with 3.
-   To perform a loop over all primes <= B1, do the following:
+   To perform a loop over all primes <= B1, do the following
+   (compile this file with -DMAIN to count primes):
 
       for (p = 2.0; p <= B1; p = getprime (p))
          {
             ...
          }
+
+   It is slightly less efficient (1.5 to 2 times) than Dan Bernstein's
+   primegen library (http://cr.yp.to/primegen.html), however it is
+   fast enough for our usage here.
 */
+
 double
 getprime (double pp)
 {
@@ -51,7 +56,7 @@ getprime (double pp)
   static long int len = 0; /* length of sieving table */
   static unsigned long int *moduli = NULL; /* offset for small primes */
 
-  if (pp == 0.0) /* free the tables, and reinitialize */
+  if (WANT_FREE_PRIME_TABLE(pp)) /* free the tables, and reinitialize */
     {
       offset = 0.0;
       current = -1;
@@ -79,12 +84,9 @@ getprime (double pp)
     {
       free (sieve);
       len *= 2;
-      sieve = (unsigned char *) malloc (len * sizeof(unsigned char));
-      if (sieve == NULL)
-        {
-          fprintf (stderr, "Error: not enough memory\n");
-          exit (EXIT_FAILURE);
-        }
+      sieve = (unsigned char *) malloc (len * sizeof (unsigned char));
+      /* assume this "small" malloc will not fail in normal usage */
+      ASSERT(sieve != NULL);
     }
 
   /* now enlarge small prime table if too small */
@@ -94,26 +96,17 @@ getprime (double pp)
 	  {
 	    nprimes = 1;
 	    primes = (unsigned *) malloc (nprimes * sizeof(unsigned long int));
-            if (primes == NULL)
-              {
-                fprintf (stderr, "Error: not enough memory\n");
-                exit (EXIT_FAILURE);
-              }
+	    /* assume this "small" malloc will not fail in normal usage */
+	    ASSERT(primes != NULL);
 	    moduli = (long unsigned int *) malloc (nprimes *
                                                    sizeof(unsigned long int));
-            if (moduli == NULL)
-              {
-                fprintf (stderr, "Error: not enough memory\n");
-                exit (EXIT_FAILURE);
-              }
+	    /* assume this "small" malloc will not fail in normal usage */
+	    ASSERT(moduli != NULL);
 	    len = 1;
 	    sieve = (unsigned char *) malloc(len *
                                        sizeof(unsigned char)); /* len=1 here */
-            if (sieve == NULL)
-              {
-                fprintf (stderr, "Error: not enough memory\n");
-                exit (EXIT_FAILURE);
-              }
+	    /* assume this "small" malloc will not fail in normal usage */
+	    ASSERT(sieve != NULL);
 	    offset = 5.0;
 	    sieve[0] = 1; /* corresponding to 5 */
 	    primes[0] = 3;
@@ -131,6 +124,8 @@ getprime (double pp)
                                            sizeof(unsigned long int));
 	    moduli = (unsigned long int *) realloc (moduli, nprimes *
                                                     sizeof(unsigned long int));
+	    /* assume those "small" realloc's will not fail in normal usage */
+	    ASSERT(primes != NULL && moduli != NULL);
 	    for (p = primes[i-1]; i < nprimes; i++)
 	      {
 		/* find next (odd) prime > p */
@@ -170,8 +165,28 @@ getprime (double pp)
   current = -1;
   while ((++current < len) && (sieve[current] == 0));
 
-  if (current < len)
-    return offset + 2.0 * (double) current;
-  else
-    abort ();
+  ASSERT(current < len); /* otherwise we found a prime gap >= sqrt(x) around x */
+  return offset + 2.0 * (double) current;
 }
+
+#ifdef MAIN
+int
+main (int argc, char *argv[])
+{
+  double p, B;
+  unsigned long pi = 0;
+
+  if (argc != 2)
+    {
+      fprintf (stderr, "Usage: getprime <bound>\n");
+      exit (EXIT_FAILURE);
+    }
+
+  B = atof (argv[1]);
+  
+  for (pi = 0, p = 2.0; p <= B; p = getprime (p), pi++);
+  printf ("pi(%1.0f)=%lu\n", B, pi);
+
+  return 0;
+}
+#endif
