@@ -1,54 +1,64 @@
 /* Modular multiplication.
 
-  Copyright 2002, 2003 Paul Zimmermann and Alexander Kruppa.
+  Copyright 2002, 2003, 2004, 2005 Paul Zimmermann and Alexander Kruppa.
 
-  This program is free software; you can redistribute it and/or modify it
-  under the terms of the GNU General Public License as published by the
-  Free Software Foundation; either version 2 of the License, or (at your
+  This file is part of the ECM Library.
+
+  The ECM Library is free software; you can redistribute it and/or modify
+  it under the terms of the GNU Lesser General Public License as published by
+  the Free Software Foundation; either version 2.1 of the License, or (at your
   option) any later version.
 
-  This program is distributed in the hope that it will be useful, but WITHOUT
-  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
-  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
-  more details.
+  The ECM Library is distributed in the hope that it will be useful, but
+  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
+  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
+  License for more details.
 
-  You should have received a copy of the GNU General Public License along
-  with this program; see the file COPYING.  If not, write to the Free
-  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
-  02111-1307, USA.
+  You should have received a copy of the GNU Lesser General Public License
+  along with the ECM Library; see the file COPYING.LIB.  If not, write to
+  the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
+  MA 02111-1307, USA.
 */
 
 #include <stdlib.h>
 #include <stdio.h>
 #include "gmp.h"
+
+/* For ECM_ERROR */
 #include "ecm.h"
 
+#include "ecm-gmp.h"
+#include "ecm-impl.h"
+
+FILE *ECM_STDOUT, *ECM_STDERR; /* define them here since needed in tune.c */
+
 /* define WANT_ASSERT to check normalization of residues */
 /* #define WANT_ASSERT 1 */
-
-#ifdef WANT_GMP_IMPL
-#include "gmp-impl.h"
-#else
-#include "ecm-gmp.h"
-#endif /* WANT_GMP_IMPL */
+/* #define DEBUG */
 
 #define ASSERT_NORMALIZED(x) ASSERT ((modulus->repr != MOD_MODMULN && \
 				      modulus->repr != MOD_REDC) || \
 			     mpz_size (x) <= mpz_size (modulus->orig_modulus))
 #define MPZ_NORMALIZED(x)    ASSERT (PTR(x)[ABSIZ(x)-1] != 0)
 
-#ifndef MOD_PLAIN_TO_REDC_THRESHOLD
-#define MOD_PLAIN_TO_REDC_THRESHOLD 20000
+#ifndef MPZMOD_THRESHOLD
+#define MPZMOD_THRESHOLD MPZMOD_THRESHOLD_DEFAULT
 #endif
 
-/* #define DEBUG */
+#ifndef REDC_THRESHOLD
+#define REDC_THRESHOLD REDC_THRESHOLD_DEFAULT
+#endif
+
+
+#ifndef GMP_NUMB_BITS
+#define GMP_NUMB_BITS __GMP_BITS_PER_MP_LIMB
+#endif
 
 void base2mod (mpres_t, mpres_t, mpres_t, mpmod_t);
+void base2mod_1 (mpres_t, mpres_t, mpmod_t);
 void REDC (mpres_t, mpres_t, mpz_t, mpmod_t);
-void mpn_REDC (mp_ptr, mp_srcptr, mp_srcptr, mp_srcptr, mp_size_t);
 void mod_mul2exp (mpz_t, unsigned int, mpmod_t);
 void mod_div2exp (mpz_t, unsigned int, mpmod_t);
-void mpz_mod_n (mpz_t, mpz_t, mpmod_t);
 
 /* returns +/-l if n is a factor of N = 2^l +/- 1 with N <= n^threshold, 
    0 otherwise.
@@ -89,6 +99,7 @@ isbase2 (mpz_t n, double threshold)
   mpz_clear (w);
   if (abs (res) > (int) (threshold * (double) lo)) 
     res = 0;
+
   return res;
 }
 
@@ -96,13 +107,99 @@ isbase2 (mpz_t n, double threshold)
 void
 base2mod (mpres_t R, mpres_t S, mpres_t t, mpmod_t modulus)
 {
-  mpz_tdiv_q_2exp (R, S, abs (modulus->bits));
-  mpz_tdiv_r_2exp (t, S, abs (modulus->bits));
+  unsigned long absbits = abs (modulus->bits);
+
+  ASSERT (R != S && R != t);
+  mpz_tdiv_q_2exp (R, S, absbits);
+  mpz_tdiv_r_2exp (t, S, absbits);
   if (modulus->bits < 0)
     mpz_add (R, R, t);
   else
     mpz_sub (R, t, R);
-   mpz_mod (R, R, modulus->orig_modulus);
+
+  /* mpz_mod (R, R, modulus->orig_modulus); */
+  while (mpz_sizeinbase (R, 2) > absbits)
+    {
+      mpz_tdiv_q_2exp (t, R, absbits);
+      mpz_tdiv_r_2exp (R, R, absbits);
+      if (modulus->bits < 0)
+        mpz_add (R, R, t);
+      else
+        mpz_sub (R, R, t);
+    }
+}
+
+/* Same, but source and result in same variable */
+void
+base2mod_1 (mpres_t RS, mpres_t t, mpmod_t modulus)
+{
+  unsigned long absbits = abs (modulus->bits);
+
+  ASSERT (RS != t);
+  while (mpz_sizeinbase (RS, 2) > absbits)
+    {
+      mpz_tdiv_q_2exp (t, RS, absbits);
+      mpz_tdiv_r_2exp (RS, RS, absbits); /* Just a truncate */
+      if (modulus->bits < 0)
+        mpz_add (RS, RS, t);
+      else
+        mpz_sub (RS, RS, t);
+    }
+}
+
+/* Fermat-mod */
+static void
+base2mod_2 (mpres_t RS, mp_size_t n, mpz_t modulus)
+{
+  mp_size_t s;
+
+  s = ABSIZ(RS);
+  if (s > n)
+    {
+      if (s == n + 1)
+        {
+          mp_ptr rp = PTR(RS);
+
+          if ((rp[n] = mpn_sub_1 (rp, rp, n, rp[n])))
+            rp[n] = mpn_add_1 (rp, rp, n, rp[n]);
+          MPN_NORMALIZE(rp, s);
+          SIZ(RS) = (SIZ(RS) > 0) ? s : -s;
+        }
+      else /* should happen rarely */
+        mpz_mod (RS, RS, modulus);
+    }
+}
+
+/* subquadratic REDC, at mpn level.
+   {orig,n} is the original modulus.
+   {aux,n} is the auxiliary modulus.
+   Requires ABSIZ(x) = 2n and ABSIZ(orig_modulus)=ABSIZ(aux_modulus)=n.
+ */
+static void
+ecm_redc_n (mp_ptr rp, mp_srcptr xp, mp_srcptr orig, mp_srcptr aux, mp_size_t n)
+{
+  mp_ptr tp, up;
+  mp_size_t nn = n + n;
+  mp_limb_t cy;
+  TMP_DECL(marker);
+
+  TMP_MARK(marker);
+  up = TMP_ALLOC_LIMBS(nn + nn);
+  ecm_mul_lo_n (up, xp, aux, n);
+  tp = up + nn;
+  mpn_mul_n (tp, up, orig, n);
+  /* add {x, 2n} and {tp, 2n}. We know that {tp, n} + {xp, n} will give
+     either 0, or a carry out. If xp[n-1] <> 0, then there is a carry. */
+#ifdef HAVE_NATIVE_mpn_add_nc
+  cy = __gmpn_add_nc (rp, tp + n, xp + n, n, (mp_limb_t) ((xp[n - 1]) ? 1 : 0));
+#else
+  cy = mpn_add_n (rp, tp + n, xp + n, n);
+  cy += mpn_add_1 (rp, rp, n, (mp_limb_t) ((xp[n - 1]) ? 1 : 0));
+#endif
+  if (cy || mpn_cmp (rp, orig, n) > 0)
+    cy -= mpn_sub_n (rp, rp, orig, n);
+  /* ASSERT ((cy == 0) && (mpn_cmp (rp, orig, n) < 0)); */
+  TMP_FREE(marker);
 }
 
 /* REDC. x and t must not be identical, t has limb growth */
@@ -119,7 +216,7 @@ REDC (mpres_t r, mpres_t x, mpz_t t, mpmod_t modulus)
       if (ALLOC(r) < n)
 	_mpz_realloc (r, n);
       rp = PTR(r);
-      mpn_REDC (rp, PTR(x), PTR(modulus->orig_modulus), 
+      ecm_redc_n (rp, PTR(x), PTR(modulus->orig_modulus), 
 		PTR(modulus->aux_modulus), n);
       MPN_NORMALIZE(rp, n);
       SIZ(r) = (SIZ(x) > 0) ? n : -n;
@@ -139,38 +236,6 @@ REDC (mpres_t r, mpres_t x, mpz_t t, mpmod_t modulus)
   ASSERT (ABSIZ(r) <= n);
 }
 
-/* subquadratic REDC, at mpn level.
-   {orig,n} is the original modulus.
-   {aux,n} is the auxiliary modulus.
-   Requires ABSIZ(x) = 2n and ABSIZ(orig_modulus)=ABSIZ(aux_modulus)=n.
- */
-void
-mpn_REDC (mp_ptr rp, mp_srcptr xp, mp_srcptr orig, mp_srcptr aux, mp_size_t n)
-{
-  mp_ptr tp, up;
-  mp_size_t nn = n + n;
-  mp_limb_t cy;
-  TMP_DECL(marker);
-
-  TMP_MARK(marker);
-  up = TMP_ALLOC_LIMBS(nn + nn);
-  mpn_mul_lo_n (up, xp, aux, n);
-  tp = up + nn;
-  mpn_mul_n (tp, up, orig, n);
-  /* add {x, 2n} and {tp, 2n}. We know that {tp, n} + {xp, n} will give
-     either 0, or a carry out. If xp[n-1] <> 0, then there is a carry. */
-#ifdef HAVE_NATIVE_mpn_add_nc
-  cy = mpn_add_nc (rp, tp + n, xp + n, n, (mp_limb_t) ((xp[n - 1]) ? 1 : 0));
-#else
-  cy = mpn_add_n (rp, tp + n, xp + n, n);
-  cy += mpn_add_1 (rp, rp, n, (mp_limb_t) ((xp[n - 1]) ? 1 : 0));
-#endif
-  if (cy || mpn_cmp (rp, orig, n) > 0)
-    cy -= mpn_sub_n (rp, rp, orig, n);
-  /* ASSERT ((cy == 0) && (mpn_cmp (rp, orig, n) < 0)); */
-  TMP_FREE(marker);
-}
-
 /* multiplies c by R^k modulo n where R=2^mp_bits_per_limb 
    n is supposed odd. Does not need to be efficient. */
 void 
@@ -193,84 +258,78 @@ mod_div2exp (mpz_t c, unsigned int k, mpmod_t modulus)
   mpz_mod (c, modulus->temp1, modulus->orig_modulus);
 }
 
-/* r <- c/R^nn mod n, where n are nn limbs.
+/* r <- c/R^nn mod n, where n has nn limbs, and R=2^GMP_NUMB_BITS.
    n must be odd.
    c must have space for at least 2*nn limbs.
+   r must have space for at least n limbs.
    c and r can be the same variable.
    The data in c is clobbered.
 */
-void 
-mpz_mod_n (mpz_ptr r, mpz_ptr c, mpmod_t modulus)
+static void 
+ecm_redc_basecase (mpz_ptr r, mpz_ptr c, mpmod_t modulus)
 {
   mp_ptr rp;
   mp_ptr cp;
   mp_srcptr np;
   mp_limb_t cy;
-  mp_ptr cys; /* vector of carries */
-  mp_limb_t q;
   mp_size_t j, nn = modulus->bits / __GMP_BITS_PER_MP_LIMB;
-  TMP_DECL(marker);
 
   ASSERT(ABSIZ(c) <= 2 * nn);
-  if (ALLOC(r) < nn)
-    _mpz_realloc (r, nn);
+  ASSERT(ALLOC(r) >= nn);
   cp = PTR(c);
   rp = PTR(r);
   np = PTR(modulus->orig_modulus);
   for (j = ABSIZ(c); j < 2 * nn; j++) 
     cp[j] = 0;
-  TMP_MARK(marker);
-  cys = TMP_ALLOC_LIMBS(nn);
+#ifndef NATIVE_REDC
   for (j = 0; j < nn; j++)
     {
-      q = cp[0] * modulus->Nprim;
-      cys[j] = mpn_addmul_1 (cp, np, nn, q);
+      cp[0] = mpn_addmul_1 (cp, np, nn, cp[0] * modulus->Nprim);
       cp++;
     }
   /* add vector of carries and shift */
-  cy = mpn_add_n (rp, cp, cys, nn);
-  TMP_FREE(marker);
+  cy = mpn_add_n (rp, cp, cp - nn, nn);
+#else
+  ecm_redc3 (cp, np, nn, modulus->Nprim);
+  /* add vector of carries and shift */
+  cy = mpn_add_n (rp, cp + nn, cp, nn);
+#endif
   /* the result of Montgomery's REDC is less than 2^Nbits + N,
      thus at most one correction is enough */
-  if (cy)
-    {
-      cy -= mpn_sub_n (rp, rp, np, nn);
-      ASSERT(cy == 0);
-    }
+  if (cy != 0)
+    mpn_sub_n (rp, rp, np, nn); /* a borrow should always occur here */
   MPN_NORMALIZE (rp, nn);
   SIZ(r) = SIZ(c) < 0 ? -nn : nn;
 }
 
 /* don't use base2 if repr == -1, i.e. -nobase2 */
 void 
-mpmod_init (mpmod_t modulus, mpz_t N, int repr, int verbose)
+mpmod_init (mpmod_t modulus, mpz_t N, int repr)
 {
   int base2;
   
-  /* 1.4 is an experimental best threshold on Cunningham numbers */
-  if ((repr != -1) && (base2 = isbase2 (N, 1.4)))
+  if ((repr != -1) && (base2 = isbase2 (N, BASE2_THRESHOLD)))
     {
-      if (verbose > 1)
-	printf ("Using special division for factor of 2^%d%c1\n",
-		abs (base2), (base2 < 0) ? '-' : '+');
-      mpmod_init_BASE2 (modulus, base2, N);
+      int r;
+      outputf (OUTPUT_VERBOSE,
+	       "Using special division for factor of 2^%d%c1\n",
+	       abs (base2), (base2 < 0) ? '-' : '+');
+      r = mpmod_init_BASE2 (modulus, base2, N);
+      ASSERT (r == 0); /* error should not happen if isbase2 is correct */
     }
-  else if (mpz_size (N) < 3 * DIV_DC_THRESHOLD / 2)
+  else if (mpz_size (N) < MPZMOD_THRESHOLD)
     {
-      if (verbose > 1)
-	printf ("Using MODMULN\n");
+      outputf (OUTPUT_VERBOSE, "Using MODMULN\n");
       mpmod_init_MODMULN (modulus, N);
     }
-  else if (mpz_sizeinbase (N, 2) < MOD_PLAIN_TO_REDC_THRESHOLD)
+  else if (mpz_sizeinbase (N, 2) < REDC_THRESHOLD)
     {
-      if (verbose > 1)
-	printf ("Using mpz_mod\n");
+      outputf (OUTPUT_VERBOSE, "Using mpz_mod\n");
       mpmod_init_MPZ (modulus, N);
     }
   else
     {
-      if (verbose > 1)
-	printf("Using REDC\n");
+      outputf (OUTPUT_VERBOSE, "Using REDC\n");
       mpmod_init_REDC (modulus, N);
     }
   
@@ -293,7 +352,7 @@ mpmod_init_MPZ (mpmod_t modulus, mpz_t N)
   return;
 }
 
-void 
+int 
 mpmod_init_BASE2 (mpmod_t modulus, int base2, mpz_t N)
 {
   int Nbits;
@@ -301,13 +360,44 @@ mpmod_init_BASE2 (mpmod_t modulus, int base2, mpz_t N)
   mpz_init_set (modulus->orig_modulus, N);
   modulus->repr = MOD_BASE2;
   modulus->bits = base2;
-  
+
   Nbits = mpz_size (N) * __GMP_BITS_PER_MP_LIMB; /* Number of bits, rounded
                                                     up to full limb */
   mpz_init2 (modulus->temp1, 2 * Nbits + __GMP_BITS_PER_MP_LIMB);
   mpz_init2 (modulus->temp2, Nbits);
   
-  return;
+  mpz_set_ui (modulus->temp1, 1);
+  mpz_mul_2exp (modulus->temp1, modulus->temp1, abs (base2));
+  if (base2 < 0)
+    mpz_sub_ui (modulus->temp1, modulus->temp1, 1);
+  else
+    mpz_add_ui (modulus->temp1, modulus->temp1, 1);
+  if (!mpz_divisible_p (modulus->temp1, N))
+    {
+       outputf (OUTPUT_ERROR, "mpmod_init_BASE2: n does not divide 2^%d%c1\n",
+                abs (base2), base2 < 0 ? '-' : '+');
+       mpz_clear (modulus->temp2);
+       mpz_clear (modulus->temp1);
+       mpz_clear (modulus->orig_modulus);
+       return ECM_ERROR;
+    }
+  
+  modulus->Fermat = 0;
+  if (base2 > 0)
+    {
+      unsigned long i;
+      for (i = base2; (i & 1) == 0; i >>= 1);
+      if (i == 1)
+        {
+          modulus->Fermat = base2;
+#ifdef HAVE_GWNUM
+          if (modulus->Fermat >= GWTHRESHOLD)
+            Fgwinit (modulus->Fermat);
+#endif
+        }
+    }
+  
+  return 0;
 }
 
 void
@@ -352,9 +442,7 @@ mpmod_init_MODMULN (mpmod_t modulus, mpz_t N)
   /* compute ceil(2^bits / N) */
   mpz_cdiv_q (modulus->temp1, modulus->temp1, modulus->orig_modulus);
   mpz_mul (modulus->multiple, modulus->temp1, modulus->orig_modulus);
-  /* Now multiple is the smallest multiple of N >= 2^bits */
-
-  return;
+  /* Now multiple is the largest multiple of N >= 2^bits */
 }
 
 void 
@@ -377,13 +465,12 @@ mpmod_init_REDC (mpmod_t modulus, mpz_t N)
 
   mpz_set_ui (modulus->temp1, 1);
   mpz_mul_2exp (modulus->temp1, modulus->temp1, Nbits);
-  if (!mpz_invert (modulus->aux_modulus, N, modulus->temp1))
-    {
-      fprintf (stderr, "mpmod_init: could not invert N\n");
-      exit (EXIT_FAILURE);
-    }
+  /* since we directly check even modulus in ecm/pm1/pp1,
+     N is odd here, thus 1/N mod 2^Nbits always exist */
+  mpz_invert (modulus->aux_modulus, N, modulus->temp1);
+
   mpz_sub (modulus->aux_modulus, modulus->temp1, modulus->aux_modulus);
-  /* ensure aux_modulus has n allocated limbs, for mpn_REDC */
+  /* ensure aux_modulus has n allocated limbs, for ecm_redc_n */
   if (ABSIZ(modulus->aux_modulus) < n)
     {
       _mpz_realloc (modulus->aux_modulus, n);
@@ -409,8 +496,6 @@ mpmod_init_REDC (mpmod_t modulus, mpz_t N)
   mpz_cdiv_q (modulus->temp1, modulus->temp1, modulus->orig_modulus);
   mpz_mul (modulus->multiple, modulus->temp1, modulus->orig_modulus);
   /* Now multiple is the largest multiple of N >= 2^bits */
-
-  return;
 }
 
 void 
@@ -425,8 +510,12 @@ mpmod_clear (mpmod_t modulus)
       mpz_clear (modulus->R3);
       mpz_clear (modulus->multiple);
       if (modulus->repr == MOD_REDC)
-	mpz_clear (modulus->aux_modulus);
+        mpz_clear (modulus->aux_modulus);
     }
+#ifdef HAVE_GWNUM
+  if (modulus->Fermat >= GWTHRESHOLD)
+    Fgwclear ();
+#endif
   
   return;
 }
@@ -434,10 +523,21 @@ mpmod_clear (mpmod_t modulus)
 void 
 mpres_init (mpres_t R, mpmod_t modulus)
 {
+  /* use mpz_sizeinbase since modulus->bits may not be initialized yet */
   mpz_init2 (R, mpz_sizeinbase (modulus->orig_modulus, 2));
 }
 
-/* R <- BASE^EXP mod modulus */ 
+/* realloc R so that it has at least the same number of limbs as modulus */
+void
+mpres_realloc (mpres_t R, mpmod_t modulus)
+{
+  if (modulus->repr == MOD_MODMULN)
+    _mpz_realloc (R, modulus->bits / GMP_NUMB_BITS);
+}
+
+/* R <- BASE^EXP mod modulus.
+   Assume EXP >= 0.
+ */
 void 
 mpres_pow (mpres_t R, mpres_t BASE, mpres_t EXP, mpmod_t modulus)
 {
@@ -452,7 +552,15 @@ mpres_pow (mpres_t R, mpres_t BASE, mpres_t EXP, mpmod_t modulus)
       unsigned int expidx;
       mp_limb_t bitmask, expbits;
 
-      expidx = mpz_size (EXP) -1;           /* point at most significant limb */
+      /* case EXP=0 */
+      if (mpz_cmp_ui (EXP, 0) == 0)
+        {
+          mpres_set_ui (R, 1, modulus); /* set result to 1 */
+          ASSERT_NORMALIZED (R);
+          return;
+        }
+
+      expidx = mpz_size (EXP) - 1;         /* point at most significant limb */
       expbits = mpz_getlimbn (EXP, expidx); /* get most significant limb */
       bitmask = ((mp_limb_t) 1) << (GMP_NUMB_BITS - 1);
 
@@ -467,7 +575,7 @@ mpres_pow (mpres_t R, mpres_t BASE, mpres_t EXP, mpmod_t modulus)
 		  ASSERT_NORMALIZED (R);
                   return;
                 }
-              expidx--;
+              expidx --;
               expbits = mpz_getlimbn (EXP, expidx);
               bitmask = (mp_limb_t) 1 << (GMP_NUMB_BITS - 1);
             }
@@ -489,7 +597,7 @@ mpres_pow (mpres_t R, mpres_t BASE, mpres_t EXP, mpmod_t modulus)
                 base2mod (modulus->temp2 , modulus->temp1, modulus->temp1, modulus);
               else if (modulus->repr == MOD_MODMULN)
                 {
-                  mpz_mod_n (modulus->temp2, modulus->temp1, modulus);
+                  ecm_redc_basecase (modulus->temp2, modulus->temp1, modulus);
                 }
               else
                 REDC (modulus->temp2, modulus->temp1, modulus->temp2, modulus);
@@ -501,7 +609,7 @@ mpres_pow (mpres_t R, mpres_t BASE, mpres_t EXP, mpmod_t modulus)
                     base2mod (modulus->temp2, modulus->temp1, modulus->temp1, modulus);
                   else if (modulus->repr == MOD_MODMULN)
                     {
-                      mpz_mod_n (modulus->temp2, modulus->temp1, modulus);
+                      ecm_redc_basecase (modulus->temp2, modulus->temp1, modulus);
                     }
                   else
                     REDC (modulus->temp2, modulus->temp1, modulus->temp2, modulus);
@@ -509,7 +617,7 @@ mpres_pow (mpres_t R, mpres_t BASE, mpres_t EXP, mpmod_t modulus)
             }
           if (expidx == 0)		/* if we just processed the least */
             break;			/* significant limb, we are done */
-          expidx--;
+          expidx --;
           expbits = mpz_getlimbn (EXP, expidx);
           bitmask = (mp_limb_t) 1 << (GMP_NUMB_BITS - 1);
         }
@@ -520,6 +628,18 @@ mpres_pow (mpres_t R, mpres_t BASE, mpres_t EXP, mpmod_t modulus)
   ASSERT_NORMALIZED (R);
 }
 
+
+/* Returns 1 if S == 0 (mod modulus), 0 otherwise */
+
+int
+mpres_is_zero (mpres_t S, mpmod_t modulus)
+{
+  mpz_mod (modulus->temp1, S, modulus->orig_modulus);
+  /* For all currently implemented representations, a zero residue has zero
+     integer representation */
+  return (mpz_sgn (modulus->temp1) == 0) ? 1 : 0;
+}
+
 /* R <- BASE^EXP mod modulus */ 
 void 
 mpres_ui_pow (mpres_t R, unsigned int BASE, mpres_t EXP, mpmod_t modulus)
@@ -550,7 +670,7 @@ mpres_ui_pow (mpres_t R, unsigned int BASE, mpres_t EXP, mpmod_t modulus)
 		  ASSERT_NORMALIZED (R);
                   return;
                 }
-              expidx--;
+              expidx --;
               expbits = mpz_getlimbn (EXP, expidx);
               bitmask = (mp_limb_t) 1 << (GMP_NUMB_BITS - 1);
             }
@@ -577,7 +697,7 @@ mpres_ui_pow (mpres_t R, unsigned int BASE, mpres_t EXP, mpmod_t modulus)
                 base2mod (modulus->temp2 , modulus->temp1, modulus->temp1, modulus);
               else if (modulus->repr == MOD_MODMULN)
                 {
-                  mpz_mod_n (modulus->temp2, modulus->temp1, modulus);
+                  ecm_redc_basecase (modulus->temp2, modulus->temp1, modulus);
                 }
               else
                 REDC (modulus->temp2, modulus->temp1, modulus->temp2, modulus);
@@ -615,6 +735,35 @@ mpres_mul (mpres_t R, mpres_t S1, mpres_t S2, mpmod_t modulus)
   ASSERT_NORMALIZED (S1);
   ASSERT_NORMALIZED (S2);
 
+#ifdef HAVE_GWNUM
+  if (modulus->repr == MOD_BASE2 && modulus->Fermat >= 1024)
+    {
+      base2mod_1 (S1, modulus->temp1, modulus);
+      base2mod_1 (S2, modulus->temp1, modulus);
+      ASSERT (mpz_sizeinbase (S1, 2) <= (unsigned) abs(modulus->bits));
+      ASSERT (mpz_sizeinbase (S2, 2) <= (unsigned) abs(modulus->bits));
+      Fgwmul (R, S1, S2);
+      return;
+    }
+#elif defined(HAVE_FFT)
+  if (modulus->repr == MOD_BASE2 && modulus->Fermat >= 32768)
+    {
+      mp_size_t n = modulus->Fermat / __GMP_BITS_PER_MP_LIMB;
+      unsigned long k;
+      
+      _mpz_realloc (R, n + 1);
+      k = mpn_fft_best_k (n, S1 == S2);
+      ASSERT(mpn_fft_next_size (n, k) == n);
+      base2mod_2 (S1, n, modulus->orig_modulus);
+      base2mod_2 (S2, n, modulus->orig_modulus);
+      mpn_mul_fft (PTR(R), n, PTR(S1), ABSIZ(S1), PTR(S2), ABSIZ(S2), k);
+      n ++;
+      MPN_NORMALIZE(PTR(R), n);
+      SIZ(R) = ((SIZ(S1) ^ SIZ(S2)) >= 0) ? n : -n;
+      return;
+    }
+#endif
+
   mpz_mul (modulus->temp1, S1, S2);
 
   switch (modulus->repr)
@@ -623,7 +772,7 @@ mpres_mul (mpres_t R, mpres_t S1, mpres_t S2, mpmod_t modulus)
       base2mod (R, modulus->temp1, modulus->temp1, modulus);
       break;
     case MOD_MODMULN:
-      mpz_mod_n (R, modulus->temp1, modulus);
+      ecm_redc_basecase (R, modulus->temp1, modulus);
       break;
     case MOD_REDC:
       REDC (R, modulus->temp1, modulus->temp2, modulus);
@@ -675,7 +824,6 @@ mpres_div_2exp (mpres_t R, mpres_t S, unsigned int n, mpmod_t modulus)
         mpz_tdiv_q_2exp (R, R, 1);
 
     ASSERT_NORMALIZED (R);
-  return;
 }
 
 void
@@ -734,14 +882,6 @@ mpres_sub_ui (mpres_t R, mpres_t S, unsigned int n, mpmod_t modulus)
       mpz_sub (modulus->temp1, S, modulus->temp1);
       mpz_mod (R, modulus->temp1, modulus->orig_modulus);
     }
-#ifdef DEBUG
-  else
-    {
-      fprintf (stderr, "mpres_sub_ui: Unexpected  representation %d\n", 
-               modulus->repr);
-      exit (EXIT_FAILURE);
-    }
-#endif
   ASSERT_NORMALIZED (R);
 }
 
@@ -775,7 +915,7 @@ mpres_set_z (mpres_t R, mpz_t S, mpmod_t modulus)
     {
       mpz_mod (modulus->temp2, S, modulus->orig_modulus);
       mpz_mul (modulus->temp1, modulus->temp2, modulus->R2);
-      mpz_mod_n (R, modulus->temp1, modulus);
+      ecm_redc_basecase (R, modulus->temp1, modulus);
     }
   else if (modulus->repr == MOD_REDC)
     {
@@ -783,18 +923,10 @@ mpres_set_z (mpres_t R, mpz_t S, mpmod_t modulus)
       mpz_mul (modulus->temp1, modulus->temp2, modulus->R2);
       REDC (R, modulus->temp1, modulus->temp2, modulus);
     }
-#ifdef DEBUG
-  else
-    {
-      fprintf (stderr, "mpres_set_z: Unexpected  representation %d\n", 
-               modulus->repr);
-      exit (EXIT_FAILURE);
-    }
-#endif
   ASSERT_NORMALIZED (R);
 }
 
-/* S must not be modulus->temp1 for REDC */
+/* R and S must not be modulus->temp1 */
 void 
 mpres_get_z (mpz_t R, mpres_t S, mpmod_t modulus)
 {
@@ -806,7 +938,8 @@ mpres_get_z (mpz_t R, mpres_t S, mpmod_t modulus)
   else if (modulus->repr == MOD_MODMULN)
     {
       mpz_set (modulus->temp1, S);
-      mpz_mod_n (R, modulus->temp1, modulus);
+      _mpz_realloc (R, modulus->bits / GMP_NUMB_BITS);
+      ecm_redc_basecase (R, modulus->temp1, modulus);
     }
   else if (modulus->repr == MOD_REDC)
     {
@@ -815,7 +948,7 @@ mpres_get_z (mpz_t R, mpres_t S, mpmod_t modulus)
 #ifdef DEBUG
   else
     {
-      fprintf (stderr, "mpres_get_z: Unexpected representation %d\n", 
+      fprintf (ECM_STDERR, "mpres_get_z: Unexpected representation %d\n", 
                modulus->repr);
       exit (EXIT_FAILURE);
     }
@@ -836,21 +969,13 @@ mpres_set_ui (mpres_t R, unsigned int n, mpmod_t modulus)
       mpz_mul_2exp (modulus->temp1, modulus->temp1, modulus->bits);
       mpz_mod (R, modulus->temp1, modulus->orig_modulus);
     }
-#ifdef DEBUG
-  else
-    {
-      fprintf (stderr, "mpres_set_ui: Unexpected representation %d\n", 
-               modulus->repr);
-      exit (EXIT_FAILURE);
-    }
-#endif
   ASSERT_NORMALIZED (R);
 }
 
 /* R <- -S mod modulus. Does not need to be efficient. */
 void
-mpres_neg (mpres_t R, mpres_t S, mpmod_t modulus)
-{ /* TODO: R < modulus  assumes  0 < S */
+mpres_neg (mpres_t R, mpres_t S, ATTRIBUTE_UNUSED mpmod_t modulus)
+{
   ASSERT_NORMALIZED (S);
   mpz_neg (R, S);
   ASSERT_NORMALIZED (R);
@@ -859,11 +984,10 @@ mpres_neg (mpres_t R, mpres_t S, mpmod_t modulus)
 int 
 mpres_invert (mpres_t R, mpres_t S, mpmod_t modulus)
 {
-  int res;
   ASSERT_NORMALIZED (S);
   if (modulus->repr == MOD_PLAIN || modulus->repr == MOD_BASE2)
     {
-      res = mpz_invert (R, S, modulus->orig_modulus);
+      int res = mpz_invert (R, S, modulus->orig_modulus);
       ASSERT_NORMALIZED (R);
       return res;
     }
@@ -872,7 +996,7 @@ mpres_invert (mpres_t R, mpres_t S, mpmod_t modulus)
       if (mpz_invert (modulus->temp2, S, modulus->orig_modulus))
         {
           mpz_mul (modulus->temp1, modulus->temp2, modulus->R3);
-          mpz_mod_n (R, modulus->temp1, modulus);
+          ecm_redc_basecase (R, modulus->temp1, modulus);
 	  ASSERT_NORMALIZED (R);
           return 1;
         }
@@ -895,7 +1019,7 @@ mpres_invert (mpres_t R, mpres_t S, mpmod_t modulus)
 #ifdef DEBUG
   else
     {
-      fprintf (stderr, "mpres_invert: Unexpected representation %d\n", 
+      fprintf (ECM_STDERR, "mpres_invert: Unexpected representation %d\n", 
                modulus->repr);
       exit (EXIT_FAILURE);
     }
@@ -916,6 +1040,6 @@ mpres_gcd (mpz_t R, mpres_t S, mpmod_t modulus)
 void 
 mpres_out_str (FILE *fd, unsigned int base, mpres_t S, mpmod_t modulus)
 {
-  mpres_get_z (modulus->temp1, S, modulus);
-  mpz_out_str (fd, base, modulus->temp1);
+  mpres_get_z (modulus->temp2, S, modulus);
+  mpz_out_str (fd, base, modulus->temp2);
 }
